# -*- coding: utf-8 -*-
"""
Specific proposal methods for sampling gravitational-wave models.
"""
import logging

from ..proposal import FlowProposal
from ..proposal.augmented import AugmentedFlowProposal

from .reparameterisations import get_gw_reparameterisation

logger = logging.getLogger(__name__)


class GWFlowProposal(FlowProposal):
    """Wrapper for FlowProposal that has defaults for CBC-PE"""
    aliases = {
        'chirp_mass': ('mass', None),
        'mass_ratio': ('mass_ratio', None),
        'ra': ('sky-ra-dec', ['DEC', 'dec', 'Dec']),
        'dec': ('sky-ra-dec', ['RA', 'ra']),
        'azimuth': ('sky-az-zen', ['zenith', 'zen', 'Zen', 'Zenith']),
        'zenith': ('sky-az-zen', ['azimuth', 'az', 'Az', 'Azimuth']),
        'theta_1': ('angle-sine', None),
        'theta_2': ('angle-sine', None),
        'tilt_1': ('angle-sine', None),
        'tilt_2': ('angle-sine', None),
        'theta_jn': ('angle-sine', None),
        'iota': ('angle-sine', None),
        'phi_jl': ('angle-2pi', None),
        'phi_12': ('angle-2pi', None),
        'phase': ('angle-2pi', None),
        'psi': ('angle-pi', None),
        'geocent_time': ('time', None),
        'a_1': ('to-cartesian', None),
        'a_2': ('to-cartesian', None),
        'luminosity_distance': ('distance', None),
    }

    def get_reparameterisation(self, reparameterisation):
        """Function to get reparameterisations that checks GW defaults and
        aliases
        """
        return get_gw_reparameterisation(reparameterisation)

    def add_default_reparameterisations(self):
        """
        Add default reparameterisations for parameters that have not been
        specified.
        """
        parameters = list(set(self.names)
                          - set(self._reparameterisation.parameters))
        logger.debug(f'Adding default reparameterisations for {parameters}')

        for p in parameters:
            logger.debug(f'Trying to add reparameterisation for {p}')
            if p in self._reparameterisation.parameters:
                logger.debug(f'Parameter {p} is already included')
                continue
            name, extra_params = self.aliases.get(p.lower(), (None, None))
            if name is None:
                logger.debug(f'{p} is not a known GW parameter')
                continue
            if extra_params is not None:
                p = [p] + [ep for ep in extra_params if ep in self.model.names]
            else:
                p = [p]
            prior_bounds = {k: self.model.bounds[k] for k in p}
            reparam, kwargs = get_gw_reparameterisation(name)
            logger.debug(
                f'Add reparameterisation for {p} with config: {kwargs}')
            self._reparameterisation.add_reparameterisation(
                reparam(parameters=p, prior_bounds=prior_bounds, **kwargs))


class AugmentedGWFlowProposal(AugmentedFlowProposal, GWFlowProposal):
    """Augmented version of GWFlowProposal.

    See :obj:`~nessai.proposal.augmented.AugmentedFlowProposal` and
    :obj:`~nessai.gw.proposal.GWFlowPropsosal`
    """
    pass



# -*- coding: utf-8 -*-
"""Test methods related to initialising and resuming the proposal method"""
import os
import pytest
from unittest import TestCase
from unittest.mock import patch, MagicMock

from nessai.proposal import FlowProposal


@pytest.mark.parametrize('kwargs',
                         [{'prior': 'uniform'},
                          {'draw_latent_kwargs': {'var': 4}}
                          ])
def test_init(model, kwargs):
    """Test init with some kwargs"""
    fp = FlowProposal(model, poolsize=1000, **kwargs)
    assert fp.model == model
    # Make sure the dummy kwargs is ignored and not added
    assert getattr(fp, 'prior', None) is None


@pytest.mark.parametrize('ef, fuzz', [(2.0, 3.0**0.5), (False, 2.0)])
@patch('nessai.flowmodel.FlowModel', new=MagicMock())
def test_initialise(tmpdir, proposal, ef, fuzz):
    """Test the initialise method"""
    p = tmpdir.mkdir('test')
    proposal.output = f'{p}/output/'
    proposal.rescaled_dims = 2
    proposal.expansion_fraction = ef
    proposal.fuzz = 2.0
    proposal.flow_config = {'model_config': {}}
    proposal.set_rescaling = MagicMock()
    proposal.verify_rescaling = MagicMock()

    FlowProposal.initialise(proposal)

    proposal.set_rescaling.assert_called_once()
    proposal.verify_rescaling.assert_called_once()
    assert proposal.populated is False
    assert proposal.initialised
    assert proposal.fuzz == fuzz
    assert os.path.exists(f'{p}/output')


def test_resume(proposal):
    """Test the resume method."""
    from numpy import array, array_equal
    proposal.initialise = MagicMock()
    proposal.mask = [1, 0]
    proposal.update_bounds = False
    proposal.weights_file = None
    FlowProposal.resume(proposal, None, {'model_config': {'kwargs': {}}})
    proposal.initialise.assert_called_once()
    assert array_equal(proposal.flow_config['model_config']['kwargs']['mask'],
                       array([1, 0]))


@patch('os.path.exists', return_value=True)
def test_resume_w_weights(proposal):
    """Test the resume method with weights"""
    proposal.initialise = MagicMock()
    proposal.flow = MagicMock()
    proposal.mask = None
    proposal.update_bounds = False
    proposal.weights_file = None
    FlowProposal.resume(proposal, None, {}, 'weights.pt')
    proposal.initialise.assert_called_once()
    proposal.flow.reload_weights.assert_called_once_with('weights.pt')


@pytest.mark.parametrize('data', [[1], None])
@pytest.mark.parametrize('count', [0, 1])
def test_resume_w_update_bounds(proposal, data, count):
    """Test the resume method with update bounds"""
    proposal.initialise = MagicMock()
    proposal.flow = MagicMock()
    proposal.mask = None
    proposal.update_bounds = True
    proposal.weights_file = None
    proposal.training_data = data
    proposal.training_count = count
    proposal.check_state = MagicMock()

    if count and data is None:
        with pytest.raises(RuntimeError) as excinfo:
            FlowProposal.resume(proposal, None, {})
        assert 'Could not resume' in str(excinfo.value)
    else:
        FlowProposal.resume(proposal, None, {})
        if data:
            proposal.check_state.assert_called_once_with(data)


@pytest.mark.parametrize('populated', [False, True])
def test_get_state(proposal, populated):
    """Test the get state method used for pickling the proposal.

    Tests cases where the proposal is and isn't populated.
    """

    proposal.populated = populated
    proposal.indices = [1, 2]
    proposal._reparameterisation = MagicMock()
    proposal.model = MagicMock()
    proposal._flow_config = {}
    proposal.pool = MagicMock()
    proposal.initialised = True
    proposal.flow = MagicMock()
    proposal.flow.weights_file = 'file'

    state = FlowProposal.__getstate__(proposal)

    assert state['resume_populated'] is populated
    assert state['pool'] is None
    assert state['initialised'] is False
    assert state['weights_file'] == 'file'
    assert '_reparameterisation' not in state
    assert 'model' not in state
    assert 'flow' not in state
    assert '_flow_config' not in state


@pytest.mark.integration_test
@pytest.mark.parametrize('reparameterisation', [False, True])
@pytest.mark.parametrize('init', [False, True])
def test_resume_pickle(model, tmpdir, reparameterisation, init):
    """Test pickling and resuming the proposal.

    Tests both with and without reparameterisations and before and after
    initialise has been called.
    """
    import pickle
    output = tmpdir.mkdir('test_integration')
    if reparameterisation:
        reparameterisations = {'default': {'parameters': model.names}}
    else:
        reparameterisations = None

    proposal = FlowProposal(model, poolsize=1000, plot=False,
                            expansion_fraction=1, output=output,
                            reparameterisations=reparameterisations)
    if init:
        proposal.initialise()

    proposal.mask = None
    proposal.resume_populated = False

    proposal_data = pickle.dumps(proposal)
    proposal_re = pickle.loads(proposal_data)
    proposal_re.resume(model, {})

    assert proposal._plot_pool == proposal_re._plot_pool
    assert proposal._plot_training == proposal_re._plot_training

    if init:
        assert proposal.fuzz == proposal_re.fuzz
        assert proposal.rescaled_names == proposal_re.rescaled_names


def test_reset(proposal):
    """Test reset method"""
    proposal.x = 1
    proposal.samples = 2
    proposal.populated = True
    proposal.populated_count = 10
    proposal._edges = {'x': 2}
    FlowProposal.reset(proposal)
    assert proposal.x is None
    assert proposal.samples is None
    assert proposal.populated is False
    assert proposal.populated_count == 0
    assert proposal.r is None
    assert proposal.alt_dist is None
    assert proposal._checked_population
    assert proposal._edges['x'] is None


@pytest.mark.integration_test
def test_reset_integration(tmpdir, model):
    """Test reset method iteration with other methods"""
    proposal = FlowProposal(model, poolsize=10)
    output = str(tmpdir.mkdir('reset_integration'))
    proposal = FlowProposal(model, output=output, plot=False,
                            poolsize=100, latent_prior='uniform_nball')

    modified_proposal = \
        FlowProposal(model, output=output, plot=False, poolsize=100,
                     latent_prior='uniform_nball')
    proposal.initialise()
    modified_proposal.initialise()

    modified_proposal.populate(model.new_point(), r=1.0)
    modified_proposal.reset()

    d1 = proposal.__getstate__()
    d2 = modified_proposal.__getstate__()
    for d in [d1, d2]:
        del d['_min']
        del d['_max']
        del d['rescale']
        del d['inverse_rescale']

    assert d1 == d2


@pytest.mark.parametrize('rescale', [True, False])
@pytest.mark.integration_test
def test_test_draw(tmpdir, model, rescale):
    """Verify that the `test_draw` method works.

    This method checks that samples can be drawn from the flow and then
    resets the flows. This test makes sure the flow is correctly reset.
    """
    output = tmpdir.mkdir('test')
    fp = FlowProposal(
        model, output=output, poolsize=100, rescale_parameters=rescale)
    fp.initialise()
    # Call these since they are worked out the first time they're called
    fp.x_dtype, fp.x_prime_dtype
    orig_state = fp.__getstate__()

    t = TestCase()
    t.maxDiff = None
    t.assertDictEqual(fp.__getstate__(), orig_state)
    fp.test_draw()
    t.assertDictEqual(fp.__getstate__(), orig_state)



# -*- coding: utf-8 -*-
"""
Integration tests for running the sampler with different configurations.
"""
import os
import torch
import pytest
import numpy as np

from nessai.flowsampler import FlowSampler
from nessai.livepoint import numpy_array_to_live_points
from nessai.model import Model


torch.set_num_threads(1)


@pytest.mark.slow_integration_test
def test_sampling_with_rescale(model, flow_config, tmpdir):
    """
    Test sampling with rescaling. Checks that flow is trained.
    """
    output = str(tmpdir.mkdir('w_rescale'))
    fp = FlowSampler(model, output=output, resume=False, nlive=100, plot=False,
                     flow_config=flow_config, training_frequency=10,
                     maximum_uninformed=9, rescale_parameters=True,
                     seed=1234, max_iteration=11, poolsize=10, max_threads=1)
    fp.run()
    assert fp.ns.proposal.flow.weights_file is not None
    assert fp.ns.proposal.training_count == 1


@pytest.mark.slow_integration_test
def test_sampling_with_inversion(model, flow_config, tmpdir):
    """
    Test sampling with inversion. Checks that flow is trained.
    """
    output = str(tmpdir.mkdir('w_rescale'))
    fp = FlowSampler(model, output=output, resume=False, nlive=100, plot=False,
                     flow_config=flow_config, training_frequency=10,
                     maximum_uninformed=9, rescale_parameters=True,
                     seed=1234, max_iteration=11, poolsize=10, max_threads=1,
                     boundary_inversion=True, update_bounds=True)
    fp.run()
    assert fp.ns.proposal.boundary_inversion == ['x', 'y']
    assert fp.ns.proposal.flow.weights_file is not None
    assert fp.ns.proposal.training_count == 1


@pytest.mark.slow_integration_test
def test_sampling_without_rescale(model, flow_config, tmpdir):
    """
    Test sampling without rescaling. Checks that flow is trained.
    """
    output = str(tmpdir.mkdir('wo_rescale'))
    fp = FlowSampler(model, output=output, resume=False, nlive=100, plot=False,
                     flow_config=flow_config, training_frequency=10,
                     maximum_uninformed=9, rescale_parameters=False, seed=1234,
                     max_iteration=11, poolsize=10)
    fp.run()
    assert fp.ns.proposal.flow.weights_file is not None
    assert fp.ns.proposal.training_count == 1


@pytest.mark.slow_integration_test
def test_sampling_with_maf(model, flow_config, tmpdir):
    """
    Test sampling with MAF. Checks that flow is trained but does not
    check convergence.
    """
    flow_config['model_config']['ftype'] = 'maf'
    output = str(tmpdir.mkdir('maf'))
    fp = FlowSampler(model, output=output, resume=False, nlive=100, plot=False,
                     flow_config=flow_config, training_frequency=10,
                     maximum_uninformed=9, rescale_parameters=True,
                     seed=1234, max_iteration=11, poolsize=10)
    fp.run()
    assert fp.ns.proposal.flow.weights_file is not None
    assert fp.ns.proposal.training_count == 1


@pytest.mark.slow_integration_test
@pytest.mark.parametrize('analytic', [False, True])
def test_sampling_uninformed(model, flow_config, tmpdir, analytic):
    """
    Test running the sampler with the two uninformed proposal methods.
    """
    output = str(tmpdir.mkdir('uninformed'))
    fp = FlowSampler(model, output=output, resume=False, nlive=100, plot=False,
                     flow_config=flow_config, training_frequency=None,
                     maximum_uninformed=10, rescale_parameters=True,
                     seed=1234, max_iteration=11, poolsize=10,
                     analytic_proposal=analytic)
    fp.run()


@pytest.mark.slow_integration_test
def test_sampling_with_n_pool(model, flow_config, tmpdir):
    """
    Test running the sampler with multiprocessing.
    """
    output = str(tmpdir.mkdir('pool'))
    fp = FlowSampler(model, output=output, resume=False, nlive=100, plot=False,
                     flow_config=flow_config, training_frequency=10,
                     maximum_uninformed=9, rescale_parameters=True,
                     seed=1234, max_iteration=11, poolsize=10, max_threads=3,
                     n_pool=2)
    fp.run()
    assert fp.ns.proposal.flow.weights_file is not None
    assert fp.ns.proposal.training_count == 1
    assert os.path.exists(output + '/result.json')


@pytest.mark.slow_integration_test
def test_sampling_resume(model, flow_config, tmpdir):
    """
    Test resuming the sampler.
    """
    output = str(tmpdir.mkdir('resume'))
    fp = FlowSampler(model, output=output, resume=True, nlive=100, plot=False,
                     flow_config=flow_config, training_frequency=10,
                     maximum_uninformed=9, rescale_parameters=True,
                     seed=1234, max_iteration=11, poolsize=10)
    fp.run()
    assert os.path.exists(os.path.join(output, 'nested_sampler_resume.pkl'))

    fp = FlowSampler(model, output=output, resume=True,
                     flow_config=flow_config)
    assert fp.ns.iteration == 11
    fp.ns.max_iteration = 21
    fp.run()
    assert fp.ns.iteration == 21
    assert os.path.exists(
        os.path.join(output, 'nested_sampler_resume.pkl.old'))


@pytest.mark.slow_integration_test
def test_sampling_resume_no_max_uninformed(model, flow_config, tmpdir):
    """
    Test resuming the sampler when there is no maximum iteration for
    the uinformed sampling.

    This test makes sure the correct proposal is loaded after resuming
    and re-initialising the sampler.
    """
    output = str(tmpdir.mkdir('resume'))
    fp = FlowSampler(model, output=output, resume=True, nlive=100, plot=False,
                     flow_config=flow_config, training_frequency=10,
                     maximum_uninformed=9, rescale_parameters=True,
                     seed=1234, max_iteration=11, poolsize=10)
    fp.run()
    assert os.path.exists(os.path.join(output, 'nested_sampler_resume.pkl'))

    fp = FlowSampler(model, output=output, resume=True,
                     flow_config=flow_config)
    assert fp.ns.iteration == 11
    fp.ns.maximum_uninformed = np.inf
    fp.ns.initialise()
    assert fp.ns.proposal is fp.ns._flow_proposal
    fp.ns.max_iteration = 21
    fp.run()
    assert fp.ns.iteration == 21
    assert os.path.exists(
        os.path.join(output, 'nested_sampler_resume.pkl.old'))


@pytest.mark.slow_integration_test
def test_sampling_with_infinite_prior_bounds(tmpdir):
    """
    Make sure the sampler runs when sampling a parameter with infinite prior \
        bounds.
    """
    from scipy.stats import norm
    output = str(tmpdir.mkdir('infinite_bounds'))

    class TestModel(Model):

        names = ['x', 'y']
        bounds = {'x': [0, 1], 'y': [-np.inf, np.inf]}
        reparameterisations = {'x': 'default', 'y': None}

        def new_point(self, N=1):
            x = np.concatenate([
                np.random.rand(N, 1),
                np.random.randn(N, 1)
            ], axis=1)
            return numpy_array_to_live_points(x, self.names)

        def log_prior(self, x):
            log_p = np.log(self.in_bounds(x))
            log_p += norm.logpdf(x['y'])
            return log_p

        def log_likelihood(self, x):
            log_l = np.zeros(x.size)
            for n in self.names:
                log_l += norm.logpdf(x[n])
            return log_l

    fs = FlowSampler(
        TestModel(),
        output=output,
        nlive=500,
        plot=False,
        proposal_plots=False
    )
    fs.run(plot=False)
    assert fs.ns.condition <= 0.1



# -*- coding: utf-8 -*-
"""
Test the AnglePair reparmeterisation
"""
import numpy as np
import pytest
from unittest.mock import MagicMock, create_autospec

from nessai.reparameterisations import AnglePair
from nessai.livepoint import get_dtype, parameters_to_live_point

angle_pairs = [
    (['ra', 'dec'], [[0, 2 * np.pi], [-np.pi / 2, np.pi / 2]]),
    (['dec', 'ra'], [[0, 2 * np.pi], [-np.pi / 2, np.pi / 2]]),
    (['ra', 'dec'], [[-np.pi, np.pi], [-np.pi / 2, np.pi / 2]]),
    (['az', 'zen'], [[0, 2 * np.pi], [0, np.pi]]),
    (['zen', 'az'], [[0, np.pi], [0, 2 * np.pi]])
]


@pytest.fixture(params=angle_pairs, scope='function')
def angles(request):
    return request.param


@pytest.fixture(scope='function')
def assert_invertibility():
    def test_invertibility(reparam, angles, radial=None):

        n = list(angles.values())[0].size
        x = np.zeros([n], dtype=get_dtype(reparam.parameters))
        x_prime = np.zeros([n], dtype=get_dtype(reparam.prime_parameters))
        log_j = 0

        for a in reparam.angles:
            x[a] = angles[a]
        if radial is not None:
            x[reparam.radial] = radial

        assert x.size == x_prime.size

        x_re, x_prime_re, log_j_re = reparam.reparameterise(
            x, x_prime, log_j)

        for a in reparam.angles:
            np.testing.assert_array_equal(x[a], x_re[a])
        if radial is not None:
            np.testing.assert_array_equal(x[reparam.radial],
                                          x_re[reparam.radial])

        x_in = np.zeros([n], dtype=get_dtype(reparam.parameters))

        x_inv, x_prime_inv, log_j_inv = \
            reparam.inverse_reparameterise(x_in, x_prime_re, log_j)

        for a in reparam.angles:
            np.testing.assert_array_almost_equal(x[a], x_inv[a])
        if radial is not None:
            np.testing.assert_array_almost_equal(x[reparam.radial],
                                                 x_inv[reparam.radial])

        np.testing.assert_array_equal(x_prime_re, x_prime_inv)
        np.testing.assert_array_almost_equal(log_j_re, -log_j_inv)

        return True

    return test_invertibility


@pytest.mark.parametrize('parameters', ['x', ['w', 'x', 'y', 'z']])
def test_parameters_error(parameters):
    """
    Make sure reparameterisations fails with too many or too few parameters.
    """
    with pytest.raises(RuntimeError) as excinfo:
        AnglePair(parameters=parameters, prior_bounds=None)
    assert 'Must use a pair' in str(excinfo.value)


def test_two_angles(angles):
    """Test the reparmaterisation with just the angles"""
    parameters = angles[0]
    prior_bounds = {parameters[0]: angles[1][0],
                    parameters[1]: angles[1][1]}

    reparam = AnglePair(parameters=parameters, prior_bounds=prior_bounds)

    if 'ra' in parameters:
        assert reparam.convention == 'ra-dec'

    if 'az' in parameters:
        assert reparam.convention == 'az-zen'

    # Make sure parameter[0] is always ra or azimuth
    assert np.ptp(reparam.prior_bounds[reparam.angles[0]]) == 2 * np.pi

    assert reparam.chi is not False
    assert hasattr(reparam.chi, 'rvs')
    assert reparam.has_prime_prior is False

    m = '_'.join(parameters[:2])
    assert reparam.angles == parameters[:2]
    assert reparam.radial == (m + '_radial')
    assert reparam.x == (m + '_x')
    assert reparam.y == (m + '_y')
    assert reparam.z == (m + '_z')


def test_ra_dec(assert_invertibility):
    """Test the invertibility when using RA and Dec"""
    parameters = ['ra', 'dec']
    prior_bounds = {parameters[0]: [0, 2 * np.pi],
                    parameters[1]: [-np.pi / 2, np.pi / 2]}

    reparam = AnglePair(parameters=parameters, prior_bounds=prior_bounds,
                        convention='ra-dec', prior='isotropic')

    n = 100
    angles = {'ra': np.random.uniform(*prior_bounds['ra'], n),
              'dec': np.arcsin(np.random.uniform(-1, 1, n))}
    assert assert_invertibility(reparam, angles)


def test_azimuth_zenith(assert_invertibility):
    """Test the inverbility when using azimuth and zenith"""
    parameters = ['az', 'zen']
    prior_bounds = {parameters[0]: [0, 2 * np.pi],
                    parameters[1]: [0, np.pi]}

    reparam = AnglePair(parameters=parameters, prior_bounds=prior_bounds,
                        convention='az-zen', prior='isotropic')

    n = 100
    angles = {'az': np.random.uniform(*prior_bounds['az'], n),
              'zen': np.arccos(np.random.uniform(-1, 1, n))}
    assert reparam.parameters[:2] == list(angles.keys())
    assert assert_invertibility(reparam, angles)


def test_w_radial(assert_invertibility):
    """Test the reparameterisation with a radial parameter"""
    parameters = ['r', 'ra', 'dec']
    prior_bounds = {parameters[0]: [0, 5],
                    parameters[1]: [0, 2 * np.pi],
                    parameters[2]: [-np.pi / 2, np.pi / 2]}

    reparam = AnglePair(parameters=parameters, prior_bounds=prior_bounds)

    assert reparam.parameters == ['ra', 'dec', 'r']
    assert reparam.angles == ['ra', 'dec']
    assert reparam.chi is False

    n = 100
    angles = {'ra': np.random.uniform(*prior_bounds['ra'], n),
              'dec': np.arcsin(np.random.uniform(-1, 1, n))}

    radial = np.random.uniform(*prior_bounds['r'], n)

    assert assert_invertibility(reparam, angles, radial=radial)


@pytest.mark.parametrize(
    'convention, input, expected',
    [
        ['az-zen', (1, 0, 0), (0, np.pi / 2, 1)],
        ['ra-dec', (1, 0, 0), (0, 0, 1)],
        ['az-zen', (-1, 0, 0), (np.pi, np.pi / 2, 1)],
        ['ra-dec', (-1, 0, 0), (np.pi, 0, 1)],
        ['az-zen', (0, 1, 0), (np.pi / 2, np.pi / 2, 1)],
        ['ra-dec', (0, 1, 0), (np.pi / 2, 0, 1)],
        ['az-zen', (0, -1, 0), (3 * np.pi / 2, np.pi / 2, 1)],
        ['ra-dec', (0, -1, 0), (3 * np.pi / 2, 0, 1)],
        ['az-zen', (0, 0, 1), (0, 0, 1)],
        ['ra-dec', (0, 0, 1), (0, np.pi / 2, 1)],
        ['az-zen', (0, 0, -1), (0, np.pi, 1)],
        ['ra-dec', (0, 0, -1), (0, -np.pi / 2, 1)],
        ['az-zen', (0, 0, 0), (0, 0, 0)],
        ['ra-dec', (0, 0, 0), (0, 0, 0)],
        ['az-zen', (1, 1, np.sqrt(2)), (np.pi / 4, np.pi / 4, 2)],
        ['ra-dec', (1, 1, np.sqrt(2)), (np.pi / 4, np.pi / 4, 2)],
        ['az-zen', (-1, -1, -np.sqrt(2)), (5 * np.pi / 4, 3 * np.pi / 4, 2)],
        ['ra-dec', (-1, -1, -np.sqrt(2)), (5 * np.pi / 4, -np.pi / 4, 2)],
    ]
)
def test_specific_points_x_prime_to_x_0_2pi(convention, input, expected):
    """Test specific points on a sphere.

    Order is (x, y, z) to (ra/az, dec/zen, r) using [0, 2pi] for ra/az.

    Test:
    - (+/-1, 0, 0) and all permutations
    - (0, 0, 0)
    - (+/-1, +-1, sqrt(2)) a point exactly 2 away from the origin
    """
    parameters = ['a', 'b']
    if convention == 'ra-dec':
        prior_bounds = {'a': [0, 2 * np.pi], 'b': [-np.pi / 2, np.pi / 2]}
    else:
        prior_bounds = {'a': [0, 2 * np.pi], 'b': [0, np.pi]}
    reparam = AnglePair(
        parameters=parameters,
        prior_bounds=prior_bounds,
        convention=convention,
    )

    x_prime = parameters_to_live_point(input, reparam.prime_parameters)
    x = parameters_to_live_point([0, 0, 0], reparam.parameters)
    log_j = 0

    out, _, _ = reparam.inverse_reparameterise(x, x_prime, log_j)

    np.testing.assert_equal(out[reparam.parameters[0]], expected[0])
    np.testing.assert_equal(out[reparam.parameters[1]], expected[1])
    np.testing.assert_equal(out[reparam.parameters[2]], expected[2])


@pytest.mark.parametrize(
    'convention, input, expected',
    [
        ['az-zen', (1, 0, 0), (0, np.pi / 2, 1)],
        ['ra-dec', (1, 0, 0), (0, 0, 1)],
        ['az-zen', (-1, 0, 0), (np.pi, np.pi / 2, 1)],
        ['ra-dec', (-1, 0, 0), (np.pi, 0, 1)],
        ['az-zen', (0, 1, 0), (np.pi / 2, np.pi / 2, 1)],
        ['ra-dec', (0, 1, 0), (np.pi / 2, 0, 1)],
        ['az-zen', (0, -1, 0), (- np.pi / 2, np.pi / 2, 1)],
        ['ra-dec', (0, -1, 0), (- np.pi / 2, 0, 1)],
        ['az-zen', (0, 0, 1), (0, 0, 1)],
        ['ra-dec', (0, 0, 1), (0, np.pi / 2, 1)],
        ['az-zen', (0, 0, -1), (0, np.pi, 1)],
        ['ra-dec', (0, 0, -1), (0, -np.pi / 2, 1)],
        ['az-zen', (0, 0, 0), (0, 0, 0)],
        ['ra-dec', (0, 0, 0), (0, 0, 0)],
        ['az-zen', (1, 1, np.sqrt(2)), (np.pi / 4, np.pi / 4, 2)],
        ['ra-dec', (1, 1, np.sqrt(2)), (np.pi / 4, np.pi / 4, 2)],
        ['az-zen', (-1, -1, -np.sqrt(2)), (-3 * np.pi / 4, 3 * np.pi / 4, 2)],
        ['ra-dec', (-1, -1, -np.sqrt(2)), (-3 * np.pi / 4, -np.pi / 4, 2)],
    ]
)
def test_specific_points_x_prime_to_x_pi_pi(convention, input, expected):
    """Test specific points on a sphere.

    Order is (x, y, z) to (ra/az, dec/zen, r) using [-pi, pi] for ra/az.

    Test:
    - (+/-1, 0, 0) and all permutations
    - (0, 0, 0)
    - (+/-1, +-1, sqrt(2)) a point exactly 2 away from the origin
    """
    parameters = ['a', 'b']
    if convention == 'ra-dec':
        prior_bounds = {'a': [-np.pi, np.pi], 'b': [-np.pi / 2, np.pi / 2]}
    else:
        prior_bounds = {'a': [-np.pi, np.pi], 'b': [0, np.pi]}
    reparam = AnglePair(
        parameters=parameters,
        prior_bounds=prior_bounds,
        convention=convention,
    )

    x_prime = parameters_to_live_point(input, reparam.prime_parameters)
    x = parameters_to_live_point([0, 0, 0], reparam.parameters)
    log_j = 0

    out, _, _ = reparam.inverse_reparameterise(x, x_prime, log_j)

    np.testing.assert_equal(out[reparam.parameters[0]], expected[0])
    np.testing.assert_equal(out[reparam.parameters[1]], expected[1])
    np.testing.assert_equal(out[reparam.parameters[2]], expected[2])


@pytest.mark.parametrize(
    'prior_bounds',
    [
        {'ra': [0, np.pi], 'dec':  [-np.pi / 2, np.pi / 2]},
        {'ra': [0, 2 * np.pi], 'dec':  [-np.pi, np.pi]},
    ]
)
def test_invalid_prior_ranges(prior_bounds):
    """Assert an error is raised the prior ranges are invalid"""
    with pytest.raises(ValueError) as excinfo:
        AnglePair(parameters=['ra', 'dec'], prior_bounds=prior_bounds)
    assert 'Invalid prior ranges' in str(excinfo.value)


def test_invalid_prior_bounds_az():
    """Assert an error is raised the prior bounds are invalid"""
    prior_bounds = {'az': [np.pi, 3 * np.pi], 'zen': [0, np.pi]}
    with pytest.raises(ValueError) as excinfo:
        AnglePair(parameters=['az', 'zen'], prior_bounds=prior_bounds)
    assert 'Prior bounds for az must be' in str(excinfo.value)


def test_invalid_prior_bounds_inc():
    """Assert an error is raised the prior bounds are invalid for the \
        inclination angle.
    """
    prior_bounds = {'ra': [0.0, 2 * np.pi], 'dec': [0, np.pi]}
    with pytest.raises(ValueError) as excinfo:
        AnglePair(
            parameters=['ra', 'dec'], prior_bounds=prior_bounds,
            convention='ra-dec'
        )
    assert 'Prior bounds for dec must be' in str(excinfo.value)


def test_unknown_prior():
    """Assert an unknown prior raises an error"""
    with pytest.raises(ValueError) as excinfo:
        AnglePair(
            parameters=['az', 'zen'],
            prior_bounds={'az': [0.0, 2 * np.pi], 'zen': [0.0, np.pi]},
            convention='az-zen',
            prior='uniform',
        )
    assert "Unknown prior: `uniform`. Choose from: ['isotropic', None]" \
        in str(excinfo.value)


def test_unknown_convention():
    """Assert an unknown convention raises an error"""
    with pytest.raises(ValueError) as excinfo:
        AnglePair(
            parameters=['az', 'zen'],
            prior_bounds={'az': [0.0, 2 * np.pi], 'zen': [0.0, np.pi]},
            convention='sky',
        )
    assert "Unknown convention: `sky`. Choose from: ['az-zen', 'ra-dec']" \
        in str(excinfo.value)


def test_log_prior():
    """Assert log_prior calls the log pdf of a chi distribution."""
    reparam = create_autospec(AnglePair)
    reparam.parameters = ['a', 'b', 'r']
    reparam.has_prior = True
    x = parameters_to_live_point([0, 0, 2.0], reparam.parameters)
    reparam.chi = MagicMock()
    reparam.chi.logpdf = MagicMock(return_value=1.0)
    out = AnglePair.log_prior(reparam, x)
    reparam.chi.assert_not_called()
    reparam.chi.logpdf.assert_called_once_with(x['r'])
    assert out == 1.0


def test_log_prior_radial():
    """Assert log_prior raises an error if chi is False"""
    reparam = create_autospec(AnglePair)
    reparam.chi = False
    reparam.has_prior = True
    with pytest.raises(RuntimeError) as excinfo:
        AnglePair.log_prior(reparam, 1)
    assert 'log_prior is not defined' in str(excinfo.value)


def test_log_prior_has_prior():
    """Assert log_prior raises an error if has_prior is False.

    This shouldn't be possible unless the user defines `chi`.
    """
    reparam = create_autospec(AnglePair)
    reparam.chi = True
    reparam.has_prior = False
    with pytest.raises(RuntimeError) as excinfo:
        AnglePair.log_prior(reparam, 1)
    assert 'log_prior is not defined' in str(excinfo.value)


def test_x_prime_log_prior():
    """Test the x prime prior"""
    reparam = create_autospec(AnglePair)
    reparam.prime_parameters = ['x', 'y', 'z']
    reparam.has_prime_prior = True
    x = parameters_to_live_point([1, 1, 1], reparam.prime_parameters)
    # Should be -1.5 * log(2pi) - 1.5
    expected = -1.5 * (1 + np.log(2 * np.pi))
    out = AnglePair.x_prime_log_prior(reparam, x)
    np.testing.assert_equal(out, expected)


def test_prime_prior_error():
    """Assert an error is raised if calling the prime prior and \
        has prime prior is false.
    """
    reparam = create_autospec(AnglePair)
    reparam.has_prime_prior = False
    with pytest.raises(RuntimeError) as excinfo:
        AnglePair.x_prime_log_prior(reparam, 1)
    assert 'x prime prior is not defined' in str(excinfo.value)



# -*- coding: utf-8 -*-
"""
Utilities for manipulating python structures such as lists and dictionaries.
"""


def replace_in_list(target_list, targets, replacements):
    """
    Replace (in place) an entry in a list with a given element.

    Parameters
    ----------
    target_list : list
        List to update
    targets : list
        List of items to update
    replacements : list
        List of replacement items
    """
    if not isinstance(targets, list):
        targets = [targets]
    if not isinstance(replacements, list):
        replacements = [replacements]

    if not len(targets) == len(replacements):
        raise RuntimeError('Targets and replacements are different lengths!')

    if not all([t in target_list for t in targets]):
        raise ValueError(f'Targets {targets} not in list: {target_list}')

    for t, r in zip(targets, replacements):
        i = target_list.index(t)
        target_list[i] = r



# -*- coding: utf-8 -*-
"""Tests related to plots in FlowProposal"""
import os
import numpy as np
import pytest
import torch
from unittest.mock import MagicMock, patch

from nessai.proposal import FlowProposal
from nessai.livepoint import numpy_array_to_live_points


@pytest.mark.parametrize('plot', [False, 'all'])
def test_training_plots(proposal, tmpdir, plot):
    """Make sure traings plots are correctly produced"""
    proposal._plot_training = plot
    output = tmpdir.mkdir('test/')

    names = ['x', 'y']
    prime_names = ['x_prime', 'y_prime']
    z = np.random.randn(10, 2)
    x = np.random.randn(10, 2)
    x_prime = x / 2
    proposal.training_data = numpy_array_to_live_points(x, names)
    proposal.training_data_prime = \
        numpy_array_to_live_points(x_prime, prime_names)
    x_gen = numpy_array_to_live_points(x, names)
    x_prime_gen = numpy_array_to_live_points(x_prime, prime_names)
    proposal.dims = 2
    proposal.rescale_parameters = names
    proposal.rescaled_names = prime_names

    proposal.forward_pass = MagicMock(return_value=(z, None))
    proposal.backward_pass = MagicMock(return_value=(x_prime_gen, np.ones(10)))
    proposal.inverse_rescale = MagicMock(return_value=(x_gen, np.ones(10)))
    proposal.check_prior_bounds = lambda *args: args
    proposal.model = MagicMock()
    proposal.model.names = names

    FlowProposal._plot_training_data(proposal, output)

    assert os.path.exists(f'{output}/x_samples.png') is bool(plot)
    assert os.path.exists(f'{output}/x_generated.png') is bool(plot)
    assert os.path.exists(f'{output}/x_prime_samples.png') is bool(plot)
    assert os.path.exists(f'{output}/x_prime_generated.png') is bool(plot)


def test_plot_pool_all(proposal):
    """Test for the plots that show the pool of samples"""
    proposal.output = 'test'
    proposal._plot_pool = 'all'
    proposal.populated_count = 0
    x = numpy_array_to_live_points(np.random.randn(10, 2), ['x', 'y'])
    with patch('nessai.proposal.flowproposal.plot_live_points') as plot:
        FlowProposal.plot_pool(proposal, None, x)
    plot.assert_called_once_with(x, c='logL', filename='test/pool_0.png')


@pytest.mark.parametrize('alt_dist', [False, True])
def test_plot_pool_1d(proposal, tmpdir, alt_dist):
    """Test `plot_pool` when plotting is not 'all'.

    Test cases when there is an alternative base distribution is used and
    when one is not used.
    """
    output = tmpdir.mkdir('test_plot_pool_1d')
    proposal.output = output
    proposal._plot_pool = True
    proposal.populated_count = 0

    z = np.random.randn(10, 2)
    x = numpy_array_to_live_points(np.random.randn(10, 2), ['x', 'y'])
    training_data = \
        numpy_array_to_live_points(np.random.randn(10, 2), ['x', 'y'])
    proposal.training_data = training_data
    log_p = torch.arange(10)

    proposal.flow = MagicMock()
    proposal.flow.device = 'cpu'
    if alt_dist:
        proposal.alt_dist = MagicMock()
        proposal.alt_dist.log_prob = MagicMock(return_value=log_p)
    else:
        proposal.flow.model.base_distribution_log_prob = \
            MagicMock(return_value=log_p)
        proposal.alt_dist = None
    with patch('nessai.proposal.flowproposal.plot_1d_comparison') as plot:
        FlowProposal.plot_pool(proposal, z, x)

    plot.assert_called_once_with(
        training_data,
        x,
        labels=['live points', 'pool'],
        filename=f'{output}/pool_0.png'
    )
    assert os.path.exists(f'{output}/pool_0_log_q.png')

    if alt_dist:
        proposal.alt_dist.log_prob.assert_called_once()
    else:
        proposal.flow.model.base_distribution_log_prob.assert_called_once()



# -*- coding: utf-8 -*-
"""Specific tests for different included flows."""
import numpy as np
import pytest
import torch

from nessai.flows import (
    RealNVP,
    NeuralSplineFlow
)


@pytest.mark.parametrize(
    'kwargs',
    [
        dict(
            net='mlp',
            batch_norm_within_layers=True,
            dropout_probability=0.5
        ),
        dict(use_volume_preserving=True),
        dict(linear_transform='permutation'),
        dict(linear_transform='svd'),
        dict(linear_transform='lu'),
        dict(linear_transform=None),
        dict(mask=np.array([[1, -1], [-1, 1]])),
        dict(mask=[1, -1]),
    ]
)
def test_with_realnvp_kwargs(kwargs):
    """Test RealNVP with specific kwargs"""
    flow = RealNVP(2, 2, 2, 2, **kwargs)
    x = torch.randn(10, 2)
    z, _ = flow.forward(x)
    assert z.shape == (10, 2)


@pytest.mark.parametrize(
    'kwargs, string',
    [
        (dict(net='res'), 'Unknown nn type: res'),
        (dict(linear_transform='test'), 'Unknown linear transform: test'),
        (dict(mask=[1, 1, -1]), 'Mask does not match number of features'),
        (dict(mask=[[-1, 1], [1, -1], [1, -1]]),
         'Mask does not match number of layers'),
    ]
)
def test_realnvp_value_errors(kwargs, string):
    """Assert incorrect values for some inputs raise an error"""
    with pytest.raises(ValueError) as excinfo:
        RealNVP(2, 2, 2, 2, **kwargs)
    assert string in str(excinfo.value)


@pytest.mark.parametrize(
    'kwargs',
    [
        dict(batch_norm_between_layers=True),
        dict(linear_transform='permutation'),
        dict(linear_transform='svd'),
        dict(linear_transform='lu'),
        dict(linear_transform=None),
        dict(num_bins=10)
    ]
)
def test_with_nsf_kwargs(kwargs):
    """Test NSF with specific kwargs"""
    flow = NeuralSplineFlow(2, 2, 2, 2, **kwargs)
    x = torch.randn(10, 2)
    z, _ = flow.forward(x)
    assert z.shape == (10, 2)



# -*- coding: utf-8 -*-
"""
Tests related to resuming.
"""
import datetime
import os
import pickle
import pytest
from unittest.mock import patch, MagicMock

from nessai.nestedsampler import NestedSampler


@pytest.fixture
def complete_sampler(model, tmpdir):
    """Complete instance of NestedSampler"""
    output = tmpdir.mkdir('output')
    ns = NestedSampler(model, output=output, poolsize=10)
    ns.initialise()
    return ns


@pytest.mark.parametrize('periodic', [False, True])
def test_checkpoint(sampler, periodic):
    """Test checkpointing method.

    Make sure a file is produced and that the sampling time is updated.
    Also checks to make sure that the iteration is recorded when periodic=False
    """
    sampler.checkpoint_iterations = [10]
    sampler.iteration = 20
    now = datetime.datetime.now()
    sampler.sampling_start_time = now
    sampler.sampling_time = datetime.timedelta()
    sampler.resume_file = 'test.pkl'

    with patch('nessai.nestedsampler.safe_file_dump') as sfd_mock:
        NestedSampler.checkpoint(sampler, periodic=periodic)

    sfd_mock.assert_called_once_with(
        sampler, sampler.resume_file, pickle, save_existing=True
    )

    assert sampler.sampling_start_time > now
    assert sampler.sampling_time.total_seconds() > 0.

    if periodic:
        assert sampler.checkpoint_iterations == [10]
    else:
        assert sampler.checkpoint_iterations == [10, 20]


def test_check_resume(sampler):
    """Test check resume method"""
    sampler.uninformed_sampling = False
    sampler.check_proposal_switch = MagicMock()
    sampler.resumed = True
    sampler._flow_proposal = MagicMock()
    sampler._flow_proposal.populated = False
    sampler._flow_proposal._resume_populated = True
    sampler._flow_proposal.indices = [1, 2, 3]

    NestedSampler.check_resume(sampler)

    sampler.check_proposal_switch.assert_called_once_with(force=True)
    assert sampler.resumed is False
    assert sampler._flow_proposal.populated is True


def test_check_resume_no_indices(sampler):
    """Test check resume method"""
    sampler.uninformed_sampling = True
    sampler.resumed = True
    sampler._flow_proposal = MagicMock()
    sampler._flow_proposal.populated = False
    sampler._flow_proposal._resume_populated = True
    sampler._flow_proposal.indices = []

    NestedSampler.check_resume(sampler)

    assert sampler.resumed is False
    assert sampler._flow_proposal.populated is False


def test_resume(model):
    """Test the resume method"""
    obj = MagicMock()
    obj.model = None
    model.likelihood_evaluations = 1
    obj._uninformed_proposal = MagicMock()
    obj._flow_proposal = MagicMock()
    obj.likelihood_evaluations = [2, 3]
    with patch('pickle.load', return_value=obj) as mock_pickle, \
         patch('builtins.open'):
        out = NestedSampler.resume('test.pkl', model)

    mock_pickle.assert_called_once()

    assert out.model == model
    assert model.likelihood_evaluations == 4
    assert obj.resumed is True


def test_get_state(sampler):
    """Test the getstate method used for pickling.

    It should remove the model.
    """
    sampler.model = MagicMock()
    state = NestedSampler.__getstate__(sampler)
    assert 'model' not in state


@pytest.mark.integration_test
def test_checkpoint_integration(complete_sampler):
    """Integration test for checkpointing the sampler."""
    complete_sampler.checkpoint()
    resume_file = os.path.join(
        complete_sampler.output, complete_sampler.resume_file)
    assert os.path.exists(resume_file)


@pytest.mark.integration_test
def test_checkpoint_resume_integration(complete_sampler, model):
    """Integration test for checkpointing the sampler."""
    complete_sampler.likelihood_evaluations = [1, 2]
    complete_sampler.checkpoint()
    resume_file = os.path.join(
        complete_sampler.output, complete_sampler.resume_file)
    assert os.path.exists(resume_file)
    ns = NestedSampler.resume(resume_file, model)
    assert ns is not None



# -*- coding: utf-8 -*-
"""
Various utilities for implementing normalising flows.
"""
import logging
import numpy as np
import torch
import torch.nn.functional as F

from nflows.transforms.normalization import BatchNorm
from nflows.transforms.lu import LULinear
from nflows.transforms.permutations import RandomPermutation
from nflows.nn.nets import MLP as NFlowsMLP

from .realnvp import RealNVP
from .maf import MaskedAutoregressiveFlow
from .nsf import NeuralSplineFlow
from .distributions import MultivariateNormal

logger = logging.getLogger(__name__)


def silu(x):
    """
    SiLU (Sigmoid-weighted Linear Unit) activation function.

    Also known as swish.

    Elfwing et al 2017: https://arxiv.org/abs/1702.03118v3
    """
    return torch.mul(x, torch.sigmoid(x))


def configure_model(config):
    """
    Setup the flow form a configuration dictionary.
    """
    kwargs = {}
    flows = {
        'realnvp': RealNVP,
        'maf': MaskedAutoregressiveFlow,
        'frealnvp': RealNVP,
        'spline': NeuralSplineFlow,
        'nsf': NeuralSplineFlow,
    }
    activations = {
        'relu': F.relu,
        'tanh': F.tanh,
        'swish': silu,
        'silu': silu
    }

    k = config.get('kwargs', None)
    if k is not None:
        if 'activation' in k and isinstance(k['activation'], str):
            try:
                k['activation'] = activations[k['activation']]
            except KeyError as e:
                raise RuntimeError(f'Unknown activation function: {e}')

        if 'var' in k and 'distribution' not in k:
            k['distribution'] = MultivariateNormal([config['n_inputs']],
                                                   var=k['var'])
            k.pop('var')

        kwargs.update(k)

    if not isinstance(config['n_inputs'], int):
        raise TypeError('Number of inputs (n_inputs) must be an int')

    fc = config.get('flow', None)
    ftype = config.get('ftype', None)
    if fc is not None:
        model = fc(config['n_inputs'], config['n_neurons'], config['n_blocks'],
                   config['n_layers'], **kwargs)
    elif ftype is not None:
        if ftype.lower() not in flows:
            raise RuntimeError(f'Unknown flow type: {ftype}. Choose from:'
                               f'{flows.keys()}')
        if ((('mask' in kwargs and kwargs['mask'] is not None) or
                ('net' in kwargs and kwargs['net'] is not None)) and
                ftype.lower() not in ['realnvp', 'frealnvp']):
            raise RuntimeError('Custom masks and networks are only '
                               'supported for RealNVP')

        model = flows[ftype.lower()](config['n_inputs'], config['n_neurons'],
                                     config['n_blocks'], config['n_layers'],
                                     **kwargs)
    else:
        raise RuntimeError("Must specify either 'flow' or 'ftype'.")

    device = torch.device(config.get('device_tag', 'cpu'))
    if device != 'cpu':
        try:
            model.to(device)
        except RuntimeError as e:
            device = torch.device('cpu')
            logger.warning(
                "Could not send the normailising flow to the "
                f"specified device {config['device']} send to CPU "
                f"instead. Error raised: {e}"
            )
    logger.debug('Flow model:')
    logger.debug(model)

    model.device = device

    return model, device


def reset_weights(module):
    """Reset parameters of a given module in place.

    Uses the ``reset_parameters`` method from ``torch.nn.Module``

    Also checks the following modules from nflows

    - nflows.transforms.normalization.BatchNorm

    Parameters
    ----------
    module : :obj:`torch.nn.Module`
        Module to reset
    """
    if hasattr(module, 'reset_parameters'):
        module.reset_parameters()
    elif isinstance(module, BatchNorm):
        # nflows BatchNorm does not have a weight reset, so must
        # be done manually
        constant = np.log(np.exp(1 - module.eps) - 1)
        module.unconstrained_weight.data.fill_(constant)
        module.bias.data.zero_()
        module.running_mean.zero_()
        module.running_var.fill_(1)
    else:
        logger.warning(f'Could not reset: {module}')


def reset_permutations(module):
    """Resets permutations and linear transforms for a given module in place.

    Resets using the original initialisation method. This needed since they
    do not have a ``reset_parameters`` method.

    Parameters
    ----------
    module : :obj:`torch.nn.Module`
        Module to reset
    """
    if isinstance(module, LULinear):
        module.cache.invalidate()
        module._initialize(identity_init=True)
    elif isinstance(module, RandomPermutation):
        module._permutation = torch.randperm(len(module._permutation))


class MLP(NFlowsMLP):
    """
    MLP which can be called with context.
    """
    def forward(self, inputs, context=None):
        """Forward method that allows for kwargs such as context.

        Parameters
        ----------
        inputs : :obj:`torch.tensor`
            Inputs to the MLP
        context : None
            Conditional inputs, must be None. Only implemented to the
            function is compatible with other methods.

        Raises
        ------
        RuntimeError
            If the context is not None.
        """
        if context is not None:
            raise NotImplementedError(
                'MLP with conditional inputs is not implemented.'
            )
        return super().forward(inputs)



# -*- coding: utf-8 -*-
"""
Test the ToCartesian reparameterisation.
"""
import numpy as np
import pytest
from unittest.mock import create_autospec, patch

from nessai.livepoint import numpy_array_to_live_points
from nessai.reparameterisations import ToCartesian


@pytest.fixture
def reparam():
    return create_autospec(ToCartesian)


@pytest.fixture
def x():
    return numpy_array_to_live_points(np.array([[0.], [1.0], [2.0]]), ['x'])


@pytest.mark.parametrize('mode', ['duplicate', 'half', 'split'])
@pytest.mark.parametrize('scale', [0.5, 2.0])
def test_init(reparam, mode, scale):
    prior_bounds = {'x': [0.0, 1.0]}
    """Test the init method"""

    def side_effect(scale, prior_bounds=None):
        reparam.scale = scale
        reparam.parameters = ['x', 'y']
        reparam.prior_bounds = prior_bounds

    with patch('nessai.reparameterisations.Angle.__init__',
               side_effect=side_effect) as super_init:
        ToCartesian.__init__(
            reparam, mode=mode, scale=scale, prior_bounds=prior_bounds
        )
    assert reparam.mode == mode
    assert reparam._zero_bound is False
    assert reparam._k == 1.0
    super_init.assert_called_once_with(scale=scale, prior_bounds=prior_bounds)


def test_init_invalid_mode(reparam):
    """Test the init method with an invalid mode"""
    with patch('nessai.reparameterisations.Angle.__init__'), \
         pytest.raises(RuntimeError) as excinfo:
        ToCartesian.__init__(reparam, mode='double')
    assert 'Unknown mode' in str(excinfo.value)


def test_rescale_angle_split(reparam, x):
    """Test method for rescaling the 'angle'.

    For ToCartesian this is the rescaling applied to the parameter.
    """
    reparam.parameters = ['x']
    reparam.prior_bounds = {'x': [0, 2]}
    reparam.mode = 'split'
    reparam.scale = np.pi
    expected_angle = np.array([0.0, -np.pi / 2, np.pi])

    x_prime = numpy_array_to_live_points(np.zeros([3, 1]), ['x_prime'])
    log_j = np.zeros(3)

    with patch('numpy.random.choice', return_value=np.array([1])):
        angle, x_out, x_prime_out, log_j_out = ToCartesian._rescale_angle(
            reparam, x, x_prime, log_j, compute_radius=False
        )

    np.testing.assert_array_equal(angle, expected_angle)
    np.testing.assert_array_equal(x_out, x)
    np.testing.assert_array_equal(x_prime_out, x_prime)
    np.testing.assert_equal(log_j_out, -np.log(2))


@pytest.mark.parametrize(
    'args',
    [
        {'mode': 'split', 'compute_radius': True},
        {'mode': 'duplicate', 'compute_radius': False}
    ]
)
def test_rescale_angle_duplicate_or_compute_radius(reparam, x, args):
    """Test method for rescaling the 'angle'.

    For ToCartesian this is the rescaling applied to the parameter.
    """
    reparam.parameters = ['x']
    reparam.prior_bounds = {'x': [0, 2]}
    reparam.mode = args['mode']
    reparam.scale = np.pi

    expected_angle = np.array([0.0, np.pi / 2, np.pi, 0.0, -np.pi / 2, -np.pi])

    x_prime = numpy_array_to_live_points(np.zeros([3, 1]), ['x_prime'])
    log_j = np.zeros(3)

    with patch('numpy.random.choice', return_value=np.array([1])):
        angle, x_out, x_prime_out, log_j_out = ToCartesian._rescale_angle(
            reparam, x, x_prime, log_j, compute_radius=args['compute_radius']
        )

    np.testing.assert_array_equal(angle, expected_angle)
    np.testing.assert_array_equal(x_out, np.concatenate([x, x]))
    np.testing.assert_array_equal(
        x_prime_out, np.concatenate([x_prime, x_prime])
    )
    np.testing.assert_equal(log_j_out, -np.log(2))


def test_inverse_rescale_angle(reparam):
    """Test the inverse method for rescaling the 'angle'."""
    reparam.parameters = ['x']
    reparam.prior_bounds = {'x': [0, 2]}

    x = numpy_array_to_live_points(np.array([[0.], [-0.5], [1.0]]), ['x'])
    expected_x = x.copy()
    expected_x['x'] = np.array([0., 1.0, 2.0])

    x_prime = numpy_array_to_live_points(
        np.array([0., 1.0, 3.0]), ['x_prime'])
    log_j = np.zeros(3)

    x_out, x_prime_out, log_j_out = ToCartesian._inverse_rescale_angle(
        reparam, x, x_prime, log_j
    )

    np.testing.assert_array_equal(x_out, expected_x)
    np.testing.assert_array_equal(x_prime_out, x_prime)
    np.testing.assert_equal(log_j_out, np.log(2))



# -*- coding: utf-8 -*-
"""
Test the base proposal class.
"""
import datetime
import logging
import pickle
import numpy as np
import pytest
from unittest.mock import Mock, call, create_autospec, patch

from nessai.livepoint import numpy_array_to_live_points
from nessai.proposal.base import (
    Proposal,
    _initialize_global_variables,
    _log_likelihood_wrapper
)


class DummyProposal(Proposal):

    def draw(self, old_param):
        pass


@pytest.fixture
def proposal():
    return create_autospec(Proposal)


def test_init(proposal):
    """Test the init method"""
    model = Mock()
    Proposal.__init__(proposal, model, n_pool=2)
    assert proposal.model is model
    assert proposal.training_count == 0
    assert proposal._checked_population is True
    assert proposal.population_time.total_seconds() == 0
    assert proposal.logl_eval_time.total_seconds() == 0


def test_init_with_draw():
    """Assert an error is raised if `draw` is not implemented."""
    model = Mock()
    with pytest.raises(TypeError) as excinfo:
        Proposal(model)
    assert "class Proposal with abstract method" in str(excinfo.value)


def test_initialised(proposal):
    """Test the initialised property"""
    proposal._initialised = True
    val = Proposal.initialised.__get__(proposal)
    assert val is True


@pytest.mark.parametrize('val', [True, False])
def test_initialised_setter(proposal, val):
    """Test the setter for initialised."""
    Proposal.initialised.__set__(proposal, val)
    assert proposal._initialised is val


def test_initialise(proposal):
    """Test the initialise method"""
    Proposal.initialise(proposal)
    assert proposal.initialised is True


def test_configure_pool(proposal):
    """Test configuring the pool"""
    model = Mock()
    proposal.model = model
    proposal.n_pool = 4
    proposal.check_acceptance = False
    proposal.pool = None
    pool = Mock()
    with patch('multiprocessing.Pool', return_value=pool) as mock_pool:
        Proposal.configure_pool(proposal)
    proposal.check_acceptance is True
    assert proposal.pool is pool
    mock_pool.assert_called_once_with(
        processes=4,
        initializer=_initialize_global_variables,
        initargs=(model,)
    )


def test_configure_pool_none(proposal, caplog):
    """Test configuring the pool when n_pool is None"""
    caplog.set_level(logging.INFO)
    proposal.n_pool = None
    proposal.pool = None
    Proposal.configure_pool(proposal)
    assert proposal.pool is None
    assert 'n_pool is none, no multiprocessing pool' in str(caplog.text)


@pytest.mark.parametrize('code', [10, 2])
def test_close_pool(proposal, code):
    """Test closing the pool"""
    pool = Mock()
    pool.close = Mock()
    pool.terminate = Mock()
    pool.join = Mock()
    proposal.pool = pool
    Proposal.close_pool(proposal, code=code)
    pool.join.assert_called_once()
    if code == 2:
        pool.terminate.assert_called_once()
        pool.pool.assert_not_called()
    else:
        pool.close.assert_called_once()
        pool.terminate.assert_not_called()
    assert proposal.pool is None


def test_evaluate_likelihoods_pool(proposal):
    """Test evaluating the likelihood with a pool"""
    samples = numpy_array_to_live_points(np.array([[1], [2]]), ['x'])
    logL = np.array([3, 4])
    proposal.pool = Mock(side_effect=True)
    proposal.pool.map = Mock(return_value=logL)
    proposal.samples = samples
    proposal.logl_eval_time = datetime.timedelta()
    proposal.model = Mock()
    proposal.model.likelihood_evaluations = 100
    Proposal.evaluate_likelihoods(proposal)
    proposal.pool.map.assert_called_once_with(
        _log_likelihood_wrapper,
        samples
    )
    proposal.logl_eval_time.total_seconds() > 0
    assert proposal.model.likelihood_evaluations == 102
    np.testing.assert_array_equal(proposal.samples['logL'], logL)


def test_evaluate_likelihoods_no_pool(proposal):
    """Test evaluating the likelihood without a pool"""
    samples = numpy_array_to_live_points(np.array([[1], [2]]), ['x'])
    logL = np.array([3, 4])
    proposal.pool = None
    proposal.samples = samples
    proposal.logl_eval_time = datetime.timedelta()
    proposal.model = Mock()
    proposal.model.evaluate_log_likelihood = Mock(side_effect=logL)
    Proposal.evaluate_likelihoods(proposal)
    proposal.model.evaluate_log_likelihood.assert_has_calls(
        [call(samples[0]), call(samples[1])]
    )
    proposal.logl_eval_time.total_seconds() > 0
    np.testing.assert_array_equal(proposal.samples['logL'], logL)


def test_draw(proposal):
    """Assert an error is raised."""
    with pytest.raises(NotImplementedError):
        Proposal.draw(proposal, None)


def test_test_draw(proposal):
    """Test the test draw method"""
    proposal.model = Mock()
    proposal.model.new_point = Mock(return_value=1)
    proposal.model.log_prior = Mock(return_value=-1)

    new_point = numpy_array_to_live_points(np.array([[1]]), ['x'])
    new_point['logP'] = -1
    proposal.draw = Mock(return_value=new_point)
    Proposal.test_draw(proposal)

    proposal.model.new_point.assert_called_once()
    proposal.draw.assert_called_once_with(1)
    proposal.model.log_prior.assert_called_once_with(new_point)


def test_test_draw_error(proposal):
    """Test the test draw method with an incorrect prior value"""
    proposal.model = Mock()
    proposal.model.new_point = Mock(return_value=1)
    proposal.model.log_prior = Mock(return_value=-1)

    new_point = numpy_array_to_live_points(np.array([[1]]), ['x'])
    new_point['logP'] = -2
    proposal.draw = Mock(return_value=new_point)
    with pytest.raises(RuntimeError) as excinfo:
        Proposal.test_draw(proposal)
    assert 'Log prior of new point is incorrect!' in str(excinfo.value)

    proposal.model.new_point.assert_called_once()
    proposal.draw.assert_called_once_with(1)
    proposal.model.log_prior.assert_called_once_with(new_point)


def test_train(proposal, caplog):
    """Test the train method."""
    caplog.set_level(logging.INFO)
    Proposal.train(proposal, 1, plot=True)
    assert 'This proposal method cannot be trained' in str(caplog.text)


def test_resume(proposal):
    """Test the resume method."""
    model = Mock()
    Proposal.resume(proposal, model)
    assert proposal.model is model


def test_getstate(proposal):
    """Test the get state method called by pickle."""
    proposal.model = Mock()
    proposal.pool = Mock()
    state = Proposal.__getstate__(proposal)
    assert state['pool'] is None
    assert 'model' not in list(state.keys())


@pytest.mark.integration_test
def test_pickling(model):
    """Test pickling the proposal"""
    proposal = DummyProposal(model, n_pool=1)
    proposal.configure_pool()
    d = pickle.dumps(proposal)
    out = pickle.loads(d)
    proposal.close_pool()
    assert out.pool is None
    assert hasattr(out, 'model') is False



# -*- coding: utf-8 -*-
"""
Testing the plotting functions.
"""
import os
import matplotlib.pyplot as plt
import numpy as np
import pytest
from unittest.mock import patch

from nessai import plot


@pytest.fixture()
def live_points(model):
    """Set of live points"""
    return model.new_point(N=100)


@pytest.fixture()
def live_points_1(model):
    """Second set of live points"""
    return model.new_point(N=200)


@pytest.fixture()
def nested_samples(live_points):
    return np.sort(live_points, order='logL')


@pytest.mark.parametrize('bounds', [None, True])
def test_plot_live_points_bounds(live_points, bounds, model):
    """Test generating a plot for a set of live points."""
    if bounds:
        bounds = model.bounds
    fig = plot.plot_live_points(live_points, bounds=bounds)
    assert fig is not None
    plt.close()


@pytest.mark.parametrize('c', [None, 'x'])
def test_plot_live_points_hue(live_points, c, model):
    """Test generating a plot for a set of live points with a hue."""
    fig = plot.plot_live_points(live_points, c=c)
    assert fig is not None
    plt.close()


def test_plot_live_points_constant_hue(live_points):
    """Test to make sure that constant hue is handled correctly"""
    live_points['logL'] = np.ones(live_points.size)
    fig = plot.plot_live_points(live_points, c='logL')
    assert fig is not None
    plt.close()


@pytest.mark.parametrize('save', [False, True])
def test_plot_live_points_save(live_points, save, model, tmpdir):
    """Test generating a plot for a set of live points and saving it."""
    if save:
        filename = tmpdir + 'lp.png'
    else:
        filename = None
    plot.plot_live_points(live_points, filename=filename)
    plt.close()


def test_plot_live_points_1d():
    """Test generating the live points plot for one parameter"""
    live_points = np.random.randn(100).view([('x', 'f8')])
    fig = plot.plot_live_points(live_points)
    assert fig is not None
    plt.close()


def test_plot_live_points_error(live_points):
    """
    Test to make sure that an error is not raised if plotting fails
    because of issues with seaborn and gwpy.
    """
    with patch('seaborn.scatterplot', side_effect=TypeError('Mock error')):
        fig = plot.plot_live_points(live_points)
    assert fig is None


@pytest.mark.parametrize('parameters', [None, ['x', 'y']])
def test_plot_1d_comparison_unstructured(parameters):
    """Test plotting live points in arrays are not structured."""
    live_points = np.random.randn(10, 2)
    plot.plot_1d_comparison(live_points, convert_to_live_points=True,
                            parameters=parameters)
    plt.close()


def test_plot_1d_comparison_unstructured_missing_flag():
    """Test plotting live points in arrays are not structured."""
    live_points = np.random.randn(10, 2)
    with pytest.raises(RuntimeError) as excinfo:
        plot.plot_1d_comparison(live_points, convert_to_live_points=False)

    assert 'not structured array' in str(excinfo.value)


@pytest.mark.parametrize('parameters', [None, ['x', 'y']])
def test_plot_1d_comparison_parameters(parameters, live_points, live_points_1):
    """Test generating a 1d comparison plot"""
    plot.plot_1d_comparison(live_points, live_points_1, parameters=parameters)
    plt.close()


@pytest.mark.parametrize('save', [False, True])
def test_plot_1d_comparison_save(save, live_points, live_points_1, tmpdir):
    """Test generating a 1d comparison plot"""
    if save:
        filename = tmpdir + 'comp.png'
    else:
        filename = None
    plot.plot_1d_comparison(live_points, live_points_1, filename=filename)
    plt.close()


@pytest.mark.parametrize('labels', [None, True])
def test_plot_1d_comparison_labels(labels, live_points, live_points_1, model):
    """Test generating a 1d comparison plot"""
    if labels:
        labels = model.names
    plot.plot_1d_comparison(live_points, live_points_1, labels=labels)
    plt.close()


def test_plot_1d_comparison_invalid_labels_list(live_points, live_points_1):
    """Assert an error is raised if the colours list is a different length."""
    with pytest.raises(ValueError) as excinfo:
        plot.plot_1d_comparison(
            live_points, live_points_1, labels=['r', 'g', 'b']
        )
    assert 'Length of labels list must match ' in str(excinfo.value)


@pytest.mark.parametrize('bounds', [None, True])
def test_plot_1d_comparison_bounds(bounds, live_points, live_points_1, model):
    """Test generating a 1d comparison plot"""
    if bounds:
        bounds = model.bounds
    plot.plot_1d_comparison(live_points, live_points_1, bounds=bounds)
    plt.close()


def test_plot_1d_comparison_1d():
    """Test generating a 1d comparison plot with only one parameter"""
    l1 = np.random.randn(100).view([('x', 'f8')])
    l2 = np.random.randn(100).view([('x', 'f8')])
    plot.plot_1d_comparison(l1, l2)
    plt.close()


def test_plot_1d_comparison_infinite_var(live_points, caplog):
    """Test generating a 1d comparirson with a variable that is not finite."""
    live_points['y'] = np.inf * np.ones(live_points.size)
    plot.plot_1d_comparison(live_points)
    plt.close()
    assert 'No finite points for y, skipping.' in str(caplog.text)


def test_plot_1d_comparison_infinite_var_comp(live_points, live_points_1,
                                              caplog):
    """Test generating a 1d comparirson with a variable that is not finite but
    the second set of points contains finite values.
    """
    live_points['y'] = np.inf * np.ones(live_points.size)
    plot.plot_1d_comparison(live_points, live_points_1)
    plt.close()
    assert 'No finite points for y, skipping.' not in str(caplog.text)


@pytest.mark.parametrize('colours', [None, ['r', 'g']])
def test_plot_1d_comparison_colours(colours, live_points, live_points_1):
    """Test generating a 1d comparison plot with only one parameter"""
    plot.plot_1d_comparison(live_points, live_points_1, colours=colours)
    plt.close()


def test_plot_1d_comparison_more_colours(model):
    """Test generating a 1d comparirson when comparing more than 10 sets
    of live points.

    The default colour palettes all use 10 colours.
    """
    points = [model.new_point(N=10) for _ in range(12)]
    plot.plot_1d_comparison(*points)
    plt.close()


def test_plot_1d_comparison_invalid_colours_list(live_points, live_points_1):
    """Assert an error is raised if the colours list is a different length."""
    with pytest.raises(ValueError) as excinfo:
        plot.plot_1d_comparison(
            live_points, live_points_1, colours=['r', 'g', 'b']
        )
    assert 'Length of colours list must match ' in str(excinfo.value)


@pytest.mark.parametrize('plot_breakdown', [False, True])
def test_plot_indices_breakdown(plot_breakdown):
    """Test plotting insertion indices with and without breakdown"""
    nlive = 100
    indices = np.random.randint(0, nlive, 1000)
    plot.plot_indices(indices, nlive=nlive, plot_breakdown=plot_breakdown)
    plt.close()


@pytest.mark.parametrize('save', [False, True])
def test_plot_indices_save(save, tmpdir):
    """Test plotting insertion indices with and without saving"""
    nlive = 100
    indices = np.random.randint(0, nlive, 1000)
    if save:
        filename = tmpdir + 'indices.png'
    else:
        filename = None
    plot.plot_indices(indices, nlive=nlive, filename=filename)
    plt.close()


def test_plot_indices_no_indices():
    """Test to ensure plotting does not fail if indices are empty"""
    plot.plot_indices([], nlive=100)
    plt.close()


@pytest.mark.parametrize('sym', [False, True])
def test_plot_loss(sym):
    """Test function for plotting the loss"""
    epoch = 10
    if sym:
        history = dict(loss=-np.arange(epoch),
                       val_loss=-np.arange(epoch))
    else:
        history = dict(loss=np.arange(epoch), val_loss=np.arange(epoch))
    plot.plot_loss(epoch, history)
    plt.close()


@pytest.mark.parametrize('save', [False, True])
def test_plot_loss_save(save, tmpdir):
    """Test function for plotting the loss"""
    epoch = 10
    history = dict(loss=np.arange(epoch), val_loss=np.arange(epoch))
    if save:
        filename = tmpdir + 'loss.png'
    else:
        filename = None
    plot.plot_loss(epoch, history, filename=filename)
    plt.close()


@pytest.mark.parametrize('save', [False, True])
def test_trace_plot_save(nested_samples, save, tmpdir):
    """Test trace plot generation saving."""
    log_x = np.linspace(-10, 0, nested_samples.size)
    if save:
        filename = tmpdir + 'trace.png'
    else:
        filename = None
    plot.plot_trace(log_x, nested_samples, filename=filename)
    plt.close()


def test_trace_plot_1d(nested_samples):
    """Test trace plot with only one parameter"""
    log_x = np.linspace(-10, 0, nested_samples.size)
    plot.plot_trace(log_x, nested_samples[['x']])
    plt.close()


def test_trace_plot_unstructured():
    """
    Test to check that trace_plot raises an error when the nested samples
    are not a structured array.
    """
    log_x = np.linspace(-10, 0, 100)
    nested_samples = np.random.randn(log_x.size, 2)
    with pytest.raises(TypeError) as excinfo:
        plot.plot_trace(log_x, nested_samples)
    plt.close()
    assert 'structured array' in str(excinfo.value)


@pytest.mark.parametrize('labels', [None, ['x', 'y', 'logL', 'logP']])
def test_trace_plot_labels(nested_samples, labels):
    """Test trace plot generation with labels."""
    log_x = np.linspace(-10, 0, nested_samples.size)
    plot.plot_trace(log_x, nested_samples, labels=labels)
    plt.close()


def test_trace_plot_labels_error(nested_samples):
    """Test to ensure error is raised if labels are incompatible"""
    log_x = np.linspace(-10, 0, nested_samples.size)
    with pytest.raises(RuntimeError) as excinfo:
        plot.plot_trace(log_x, nested_samples, labels=['1', '2', '3'])

    assert 'Missing labels' in str(excinfo.value)


def test_histogram_plot():
    """Test the basic histogram plot"""
    x = np.random.randn(100)
    with patch('matplotlib.pyplot.hist') as mocked_hist:
        plot.plot_histogram(x, bins=10, density=True, label='test')
    mocked_hist.assert_called_once_with(
        x, bins=10, density=True, histtype='step')


def test_histogram_plot_save(tmpdir):
    """Test to make sure the figure is saved."""
    x = np.random.randn(100)
    fig = plot.plot_histogram(x, filename=tmpdir + 'hist.png')
    assert fig is None
    assert os.path.exists(tmpdir + 'hist.png')



# -*- coding: utf-8 -*-
"""
Test the base reparameterisation.
"""
import numpy as np
from numpy.testing import assert_equal
import pytest

from nessai.reparameterisations import Reparameterisation


@pytest.mark.parametrize('name', ['x1', ['x1']])
@pytest.mark.parametrize('prior_bounds', [[0, 1], (0, 1), {'x1': [0, 1]}])
def test_init(name, prior_bounds):
    """Test the init method with the allowed types of inputs"""
    reparam = Reparameterisation(parameters=name, prior_bounds=prior_bounds)
    assert reparam.parameters == ['x1']
    assert reparam.prime_parameters == ['x1_prime']
    assert_equal(reparam.prior_bounds, {'x1': np.array([0, 1])})


def test_init_infinite_bounds():
    """Test the init method with infinite prior bounds"""
    reparam = Reparameterisation(
        parameters=['x', 'y'],
        prior_bounds={'x': [0, 1], 'y': [0, np.inf]}
    )
    assert reparam.parameters == ['x', 'y']
    assert reparam.prime_parameters == ['x_prime', 'y_prime']
    assert_equal(reparam.prior_bounds['x'], [0, 1])
    assert_equal(reparam.prior_bounds['y'], [0, np.inf])


def test_infinite_bounds_error():
    """Test to ensure infinite prior bounds raise an error.

    Only applies if `requires_bounded_prior` is True.
    """
    class TestReparam(Reparameterisation):
        requires_bounded_prior = True

    with pytest.raises(RuntimeError) as excinfo:
        TestReparam(
            parameters=['x', 'y'],
            prior_bounds={'x': [0, 1], 'y': [0, np.inf]}
        )
    assert 'requires finite prior' in str(excinfo.value)


def test_no_prior_bounds():
    """Test not providing prior bounds."""
    class TestReparam(Reparameterisation):
        requires_bounded_prior = False

    reparam = TestReparam(parameters=['x', 'y'])
    assert reparam.prior_bounds is None


def test_no_prior_bounds_error():
    """Test missing prior bounds error.

    Only applies if `requires_bounded_prior` is True.
    """
    class TestReparam(Reparameterisation):
        requires_bounded_prior = True

    with pytest.raises(RuntimeError) as excinfo:
        TestReparam(parameters=['x', 'y'])
    assert 'requires prior bounds' in str(excinfo.value)


def test_parameters_error():
    with pytest.raises(TypeError) as excinfo:
        Reparameterisation(parameters={'x': [0, 1]})
    assert 'Parameters must be a str or list' in str(excinfo.value)


def test_missing_bounds():
    with pytest.raises(RuntimeError) as excinfo:
        Reparameterisation(parameters=['x', 'y'], prior_bounds={'x': [0, 1]})
    assert 'Mismatch' in str(excinfo.value)


def test_incorrect_bounds_type():
    with pytest.raises(TypeError) as excinfo:
        Reparameterisation(parameters=['x', 'y'], prior_bounds=1)
    assert 'Prior bounds must be' in str(excinfo.value)


def test_methods_not_implemented():
    """Test to ensure class fails if user does not define the methods"""
    reparam = Reparameterisation(parameters='x', prior_bounds=[0, 1])

    with pytest.raises(NotImplementedError):
        reparam.reparameterise(None, None, None)

    with pytest.raises(NotImplementedError):
        reparam.inverse_reparameterise(None, None, None)



# -*- coding: utf-8 -*-
"""
General tests of `get_reparameterisation`.
"""
import pytest

from nessai.reparameterisations import (
    Angle,
    AnglePair,
    NullReparameterisation,
    Reparameterisation,
    Rescale,
    RescaleToBounds,
    ToCartesian,
    get_reparameterisation
)

# List of known reparameterisations, the class and expected kwargs.
known_reparameteristions = [
    ('default', RescaleToBounds, {}),
    ('rescaletobounds', RescaleToBounds, {}),
    ('rescale-to-bounds', RescaleToBounds, {}),
    ('offset', RescaleToBounds, {'offset': True}),
    (
        'inversion', RescaleToBounds,
        {'detect_edges': True, 'boundary_inversion': True,
         'inversion_type': 'split'}
    ),
    (
        'inversion-duplicate', RescaleToBounds,
        {'detect_edges': True, 'boundary_inversion': True,
         'inversion_type': 'duplicate'}
    ),
    (
        'logit', RescaleToBounds,
        {'update_bounds': False, 'rescale_bounds': [0.0, 1.0],
         'post_rescaling': 'logit'}
    ),
    ('scale', Rescale, {}),
    ('rescale', Rescale, {}),
    ('angle', Angle, {}),
    ('angle-pi', Angle, {'scale': 2.0, 'prior': 'uniform'}),
    ('angle-2pi', Angle, {'scale': 1.0, 'prior': 'uniform'}),
    ('angle-sine', Angle, {'scale': 1.0, 'prior': 'sine'}),
    ('angle-pair', AnglePair, {}),
    ('to-cartesian', ToCartesian, {}),
    ('none', NullReparameterisation, {}),
    ('null', NullReparameterisation, {}),
    (None, NullReparameterisation, {})
]


@pytest.mark.parametrize(
    'inputs',
    known_reparameteristions,
)
def test_get_reparameterisation(inputs):
    """Test all of the included reparameterisations."""
    name, expected_class, expected_kwargs = inputs
    reparam, kwargs = get_reparameterisation(name)
    assert reparam is expected_class
    assert kwargs == expected_kwargs


def test_get_reparameterisation_with_class():
    """Test the case of class that inherits from Reparameterisation"""
    class TestReparam(Reparameterisation):
        pass
    reparam, kwargs = get_reparameterisation(TestReparam)
    assert reparam is TestReparam
    assert not kwargs


def test_get_reparameterisation_with_class_error():
    """Test the case of class that doe not inherit from Reparameterisation.

    This should raise an error.
    """
    class TestReparam:
        pass
    with pytest.raises(TypeError) as excinfo:
        get_reparameterisation(TestReparam)
    assert 'Reparameterisation must be' in str(excinfo.value)


def test_get_reparameterisation_unknown_name():
    """Assert an error is raised in the string is not recognised."""
    with pytest.raises(ValueError) as excinfo:
        get_reparameterisation('shift')
    assert 'Unknown reparameterisation: shift' in str(excinfo.value)


def test_get_reparameterisation_invalid_input():
    """Assert an error is raised if the input type is invalid."""
    with pytest.raises(TypeError) as excinfo:
        get_reparameterisation(2)
    assert 'Reparameterisation must be' in str(excinfo.value)



# -*- coding: utf-8 -*-
"""Test the augment proposal"""
import numpy as np
import pytest
from unittest.mock import create_autospec, MagicMock, patch

from nessai.livepoint import numpy_array_to_live_points
from nessai.proposal import AugmentedFlowProposal


@pytest.fixture
def proposal():
    return create_autospec(AugmentedFlowProposal)


@pytest.fixture
def x():
    _x = np.concatenate([np.random.rand(10, 2), np.random.randn(10, 2)],
                        axis=1)
    return numpy_array_to_live_points(_x, ['x', 'y', 'e_1', 'e_2'])


def test_init(model):
    """Test the init function"""
    AugmentedFlowProposal(model, poolsize=100)


@patch('nessai.proposal.augmented.FlowModel')
@patch('os.makedirs')
@patch('os.path.exists', return_value=False)
def test_initialise(mock_osexists, mock_makedirs, mock_flow, proposal):
    """Test the method the initialise the proposal"""
    proposal.output = 'output'
    proposal.rescaled_dims = 4
    proposal.augment_dims = 2
    proposal.expansion_fraction = 1.0
    proposal.flow_config = {'model_config': {}}
    proposal.set_rescaling = MagicMock()
    proposal.verify_rescaling = MagicMock()
    AugmentedFlowProposal.initialise(proposal)

    assert proposal.initialised is True
    assert proposal.fuzz == (2 ** 0.25)
    assert proposal.flow_config['model_config']['kwargs']['mask'].tolist() == \
        [1, 1, -1, -1]
    proposal.set_rescaling.assert_called_once()
    proposal.verify_rescaling.assert_called_once()
    mock_flow.assert_called_once()
    proposal.flow.initialise.assert_called_once()
    mock_osexists.assert_called_once_with('output')
    mock_makedirs.assert_called_once_with('output', exist_ok=True)


@patch('nessai.proposal.FlowProposal.set_rescaling')
def test_set_rescaling(mock, proposal):
    """Test the set rescaling method"""
    proposal.names = ['x', 'y']
    proposal.rescaled_names = ['x_prime', 'y_prime']
    proposal.rescale_parameters = proposal.names
    proposal.augment_dims = 2
    AugmentedFlowProposal.set_rescaling(proposal)

    assert proposal.names == ['x', 'y', 'e_0', 'e_1']
    assert proposal.rescaled_names == ['x_prime', 'y_prime', 'e_0', 'e_1']
    assert proposal.augment_names == ['e_0', 'e_1']
    mock.assert_called_once()


@pytest.mark.parametrize('generate', ['zeroes', 'gaussian'])
@patch('numpy.random.randn', return_value=np.ones(10))
@patch('numpy.zeros', return_value=np.ones(10))
def test_rescaling(mock_zeros, mock_randn, proposal, x, generate):
    """Test the rescaling method"""
    proposal._base_rescale = MagicMock(return_value=[x, np.ones(x.size)])
    proposal.augment_names = ['e_1']
    proposal.augment_dims = 1

    AugmentedFlowProposal._augmented_rescale(
        proposal, x, generate_augment=generate, test=True)

    proposal._base_rescale.assert_called_once_with(
        x, compute_radius=False, test=True)

    if generate == 'zeroes':
        mock_zeros.assert_called_once_with(x.size)
    else:
        mock_randn.assert_called_once_with(x.size)


@pytest.mark.parametrize('compute_radius', [False, True])
@patch('numpy.random.randn', return_value=np.arange(10))
@patch('numpy.zeros', return_value=np.ones(10))
def test_rescaling_generate_none(mock_zeros, mock_randn, proposal, x,
                                 compute_radius):
    """Test the rescaling method with generate_augment=None"""
    proposal._base_rescale = MagicMock(return_value=[x, np.ones(x.size)])
    proposal.augment_names = ['e_1']
    proposal.augment_dims = 1
    proposal.generate_augment = 'zeros'

    AugmentedFlowProposal._augmented_rescale(
        proposal, x, generate_augment=None, test=True,
        compute_radius=compute_radius)

    proposal._base_rescale.assert_called_once_with(
        x, compute_radius=compute_radius, test=True)

    if not compute_radius:
        mock_randn.assert_called_once_with(x.size)
    else:
        mock_zeros.assert_called_once_with(x.size)


def test_rescaling_generate_unknown(proposal, x):
    """Test the rescaling method with with an unknown method for generate"""
    proposal._base_rescale = MagicMock(return_value=[x, np.ones(x.size)])

    with pytest.raises(RuntimeError) as excinfo:
        AugmentedFlowProposal._augmented_rescale(
            proposal, x, generate_augment='ones')
    assert 'Unknown method' in str(excinfo.value)


@pytest.mark.parametrize('marg', [False, True])
@patch('scipy.stats.norm.logpdf')
def test_augmented_prior(mock, marg, proposal, x):
    """Test the augmented prior with and without marginalistion"""
    proposal.marginalise_augment = marg
    proposal.augment_names = ['e_1', 'e_2']
    log_prior = AugmentedFlowProposal.augmented_prior(proposal, x)
    if marg:
        assert log_prior == 0
    else:
        np.testing.assert_array_equal(x['e_1'], mock.call_args_list[0][0][0])
        np.testing.assert_array_equal(x['e_2'], mock.call_args_list[1][0][0])


@patch('nessai.proposal.flowproposal.FlowProposal.log_prior', return_value=1)
def test_log_prior(mock_prior, proposal, x):
    """Test the complete log prior"""
    proposal.augmented_prior = MagicMock(return_value=1)

    log_p = AugmentedFlowProposal.log_prior(proposal, x)

    mock_prior.assert_called_once_with(x)
    proposal.augmented_prior.assert_called_once_with(x)
    assert log_p == 2


@patch('nessai.proposal.flowproposal.FlowProposal.x_prime_log_prior',
       return_value=1)
def test_prime_log_prior(mock_prior, proposal, x):
    """Test the complete prime log prior"""
    proposal.augmented_prior = MagicMock(return_value=1)

    log_p = AugmentedFlowProposal.x_prime_log_prior(proposal, x)

    mock_prior.assert_called_once_with(x)
    proposal.augmented_prior.assert_called_once_with(x)
    assert log_p == 2


def test_marginalise_augment(proposal):
    """Test the marginalise augment function"""
    proposal.n_marg = 5
    proposal.augment_dims = 2
    x = np.concatenate([np.random.randn(3, 2), np.zeros((3, 2))], axis=1)
    z = np.random.randn(15, 4)
    log_prob = np.random.randn(15)
    proposal.flow = MagicMock()
    proposal.flow.forward_and_log_prob = \
        MagicMock(return_value=[z, log_prob])
    log_prob_out = AugmentedFlowProposal._marginalise_augment(proposal, x)

    assert len(log_prob_out) == 3


@pytest.mark.parametrize('log_p', [np.ones(2), np.array([-1, np.inf])])
@pytest.mark.parametrize('marg', [False, True])
def test_backward_pass(proposal, model, log_p, marg):
    """Test the backward pass method"""
    n = 2
    acc = int(np.isfinite(log_p).sum())
    x = np.random.randn(n, model.dims)
    z = np.random.randn(n, model.dims)
    proposal._marginalise_augment = MagicMock(return_value=log_p)
    proposal.inverse_rescale = \
        MagicMock(side_effect=lambda a: (a, np.ones(a.size)))
    proposal.rescaled_names = model.names
    proposal.alt_dist = None
    proposal.check_prior_bounds = MagicMock(side_effect=lambda a, b: (a, b))
    proposal.flow = MagicMock()
    proposal.flow.sample_and_log_prob = \
        MagicMock(return_value=[x, log_p])

    proposal.marginalise_augment = marg

    x_out, log_p = AugmentedFlowProposal.backward_pass(proposal, z)

    assert len(x_out) == acc
    proposal.inverse_rescale.assert_called_once()
    proposal.flow.sample_and_log_prob.assert_called_once_with(
        z=z, alt_dist=None)

    assert proposal._marginalise_augment.called is marg


@pytest.mark.integration_test
def test_w_default_rescaling(model, tmpdir):
    """Integration test to make sure augmented proposal works with the default
    rescaling method.
    """
    x = model.new_point(100)

    x_out = \
        2 * (x['x'] - model.bounds['x'][0]) / (np.ptp(model.bounds['x'])) - 1
    y_out = \
        2 * (x['y'] - model.bounds['y'][0]) / (np.ptp(model.bounds['y'])) - 1

    output = tmpdir.mkdir('testdir')
    proposal = AugmentedFlowProposal(model, output=output, poolsize=100,
                                     augment_dims=2, rescale_parameters=True,
                                     update_bounds=False)

    proposal.initialise()

    assert proposal.rescaled_names == ['x_prime', 'y_prime', 'e_0', 'e_1']

    x_prime, log_j = proposal.rescale(x)

    np.testing.assert_array_equal(x_out, x_prime['x_prime'])
    np.testing.assert_array_equal(y_out, x_prime['y_prime'])

    x_inv, log_j_inv = proposal.inverse_rescale(x_prime)

    np.testing.assert_array_equal(log_j, -log_j_inv)
    np.testing.assert_array_equal(x[['x', 'y']], x_inv[['x', 'y']])


@pytest.mark.integration_test
def test_w_reparameterisation(model, tmpdir):
    """Integration test to make sure augmented proposal works with
    reparameterisaitons configured using the dictionary method.
    """
    x = model.new_point(100)
    reparameterisations = {
        'x': {'reparameterisation': 'rescaletobounds', 'update_bounds': False},
        'y': {'reparameterisation': 'scale', 'scale': 2.0}
    }

    x_out = \
        2 * (x['x'] - model.bounds['x'][0]) / (np.ptp(model.bounds['x'])) - 1
    y_out = x['y'] / 2.0

    output = tmpdir.mkdir('testdir')
    proposal = AugmentedFlowProposal(model, output=output, poolsize=100,
                                     augment_dims=2,
                                     reparameterisations=reparameterisations)

    proposal.initialise()

    assert proposal.rescaled_names == ['x_prime', 'y_prime', 'e_0', 'e_1']

    x_prime, log_j = proposal.rescale(x)

    np.testing.assert_array_equal(x_out, x_prime['x_prime'])
    np.testing.assert_array_equal(y_out, x_prime['y_prime'])

    x_inv, log_j_inv = proposal.inverse_rescale(x_prime)

    np.testing.assert_array_equal(log_j, -log_j_inv)
    np.testing.assert_array_equal(x[['x', 'y']], x_inv[['x', 'y']])



#!/usr/bin/env python

# Example of using Nessai with Bilby (Requires separate installation)
# with unbounded priors. We use bilby since the priors include a sample
# method which is used to redefine `Model.new_point`. Without this we
# would have to redefine the method ourselves.

import bilby
import numpy as np

# The output from the sampler will be saved to:
# '$(outdir)/$(label)_nessai/'
# alongside the usual bilby outputs
outdir = './outdir/'
label = 'bilby_unbounded_priors'

# Setup the bilby logger
bilby.core.utils.setup_logger(outdir=outdir, label=label, log_level='INFO')

# Define a likelihood using Bilby


class SimpleGaussianLikelihood(bilby.Likelihood):

    def __init__(self):
        """
        A very simple Gaussian likelihood
        """
        super().__init__(parameters={'x': None, 'y': None})

    def log_likelihood(self):
        x = self.parameters['x']
        y = self.parameters['y']
        return -0.5*(x ** 2. + y ** 2.) - np.log(2.0 * np.pi)


# Define priors, we'll use Gaussians since they're unbounded.
priors = dict(x=bilby.core.prior.Gaussian(0, 5, 'x'),
              y=bilby.core.prior.Gaussian(0, 10, 'y'))

# Instantiate the likelihood
likelihood = SimpleGaussianLikelihood()

# Configure the normalising flow
flow_config = dict(
        max_epochs=50,
        patience=10,
        model_config=dict(n_blocks=2, n_neurons=4, n_layers=2,
                          device_tag='cpu',
                          kwargs=dict(batch_norm_between_layers=True))
        )

# Run using bilby.run_sampler, any kwargs are parsed to the sampler
# NOTE: when using Bilby if the priors can be sampled analytically  the flag
# `analytic_priors` enables faster initial sampling. See 'further details' in
# the documentation for more details
# We need to disable the rescaling since we can't rescale without bounds.
# Alternatively a different reparameterisation could be used, in this case
# we can used the 'Rescale' reparameterisation that rescales the inputs by
# a constant, this is useful when we do not prior bounds we can use.
result = bilby.run_sampler(outdir=outdir, label=label, resume=False, plot=True,
                           likelihood=likelihood, priors=priors,
                           sampler='nessai', nlive=1000,
                           maximum_uninformed=2000, flow_config=flow_config,
                           injection_parameters={'x': 0.0, 'y': 0.0},
                           reparameterisations={
                               'scale': {'parameters': ['x', 'y'],
                                         'scale': [5, 10]}},
                           proposal_plots='min',
                           analytic_priors=False, seed=1234)
